import numpy as np
from enum import IntEnum
from typing import List, Sequence, Union

from .mixins import Named, ListOfNamed


class AffineTransform(object):

    def __init__(self, transform_coord_sys: 'CoordinateSystem',
                 affine: Sequence):
        """Affine transformation from one coordinate system to another.

        The AffineTransform represents an affine transformation to a new
        stereotaxic space.

        Args:
            transform_coord_sys: The coordinate system of the data after the
                application of the affine transformation.
            affine: The affine transform as a numpy array with a shape of
                (4, 4).

        Raises:
            TypeError: If the affine is none or not convertible to a a numpy
                array of floats.
            ValueError: If the affine does not have a shape of (4, 4).

        Examples:

            Create an affine transform to scanner space.

            >>> import numpy as np
            >>> from nimesh import AffineTransform, CoordinateSystem
            >>> affine = np.eye(4)
            >>> transform = AffineTransform(CoordinateSystem.SCANNER, affine)

        """

        # The affine must be convertible to a numpy array with a shape of
        # (4, 4).
        if affine is None:
            raise TypeError('\'affine\' cannot be None.')

        try:
            affine = np.array(affine, dtype=np.float64)
        except Exception:
            raise TypeError('\'affine\' must be convertible to a numpy array '
                            'of floats.')

        if affine.ndim != 2 or affine.shape != (4, 4):
            raise ValueError('\'affine\' must have a shape of (4, 4), '
                             'not {}.'.format(affine.shape))

        self._affine = affine
        self._transform_coord_sys = transform_coord_sys

    @property
    def affine(self) -> np.array:
        """Returns the affine transform as a numpy array."""
        return self._affine.copy()

    @property
    def transform_coord_sys(self) -> 'CoordinateSystem':
        """Returns the coord. system after the application of the affine."""
        return self._transform_coord_sys


class CoordinateSystem(IntEnum):
    """The possible coordinate systems of meshes."""
    UNKNOWN = 0
    SCANNER = 1
    RAS = 2
    LPS = 3
    VOXEL = 4
    MNI = 5
    TALAIRACH = 6


class Label(Named):

    def __init__(self, name: str, color: Sequence[int] = (0, 0, 0, 0)):
        """One label of a segmentation.

        The Label class represents the data of one label of a segmentation.
        Within a segmentation, each label and each vertex are assigned a
        key creating a mapping between vertices and label data.

        Args:
            name: The name of the label, for example: left superior frontal.
        """

        super().__init__(name)

        self._color = tuple(color)

    @property
    def color(self):
        """Returns the color of the label."""
        return self._color


class Mesh(object):

    def __init__(
            self, vertices: Sequence, triangles: Sequence,
            coordinate_system: CoordinateSystem = CoordinateSystem.UNKNOWN,
            normals: Sequence = None):
        """Triangle mesh of the cortical surface.

        The Mesh class represents a polygon mesh of the cortical surface
        defined by vertices and triangles.

        Args:
            vertices: The vertices of the mesh. Must a sequence that can be
                converted to a numpy array of floats with a shape of (N, 3)
                where N is the number of vertices.
            triangles: The triangles of the mesh. Must be a sequence that
                can be converted to a numpy array of integers with a shape of
                (M, 3) where M is the number of triangles.
            coordinate_system (optional): The coordinate system of the
                vertices (see CoordinateSystem). Defaults to VOXEL.
            normals (optional): The normals of the vertices. Must be a sequence
                that can be converted to a numpy array of floats with a
                shape of (N, 3) where N is the number of vertices.

        Raises:
            TypeError: If vertices or triangles cannot be converted to numpy
                arrays.
            TypeError: If coordinate_system is not a CoordinateSystem member
                value.
            ValueError: If vertices or triangles do not have a shape of (N, 3)
                and (M, 3), respectively.

        Examples:
            Create a mesh with a single triangle.

            >>> from nimesh import Mesh
            >>> vertices = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
            >>> triangles = [[0, 1, 2]]
            >>> mesh = Mesh(vertices, triangles)
            >>> print(mesh)
            Mesh: 3 vertices, 1 triangles

        """

        # Special case of vertices or triangles set to None.
        if vertices is None or triangles is None:
            raise TypeError('\'vertices\' and \'triangles\' cannot be None')

        # Try to convert the vertices and triangles to numpy arrays.
        try:
            vertices = np.array(vertices, dtype=np.float64)
        except Exception:
            raise TypeError('\'vertices\' must be convertible to a numpy '
                            'array of floats.')

        try:
            triangles = np.array(triangles, dtype=np.int64)
        except Exception:
            raise TypeError('\'triangles\' must be convertible to a numpy '
                            'array of integers.')

        # The shape of vertices and triangles must be (?, 3).
        if vertices.ndim != 2 or vertices.shape[1] != 3:
            raise ValueError('\'vertices\' must have a shape of (N, 3) not '
                             '{}.'.format(vertices.shape))

        if triangles.ndim != 2 or triangles.shape[1] != 3:
            raise ValueError('\'triangles\' must have a shape of (M, 3) not '
                             '{}.'.format(triangles.shape))

        if not isinstance(coordinate_system, CoordinateSystem):
            raise TypeError('\'coordinate_system\' must be a member value of '
                            'CoordinateSystem, not a {}.'
                            .format(type(coordinate_system)))

        self._vertices = vertices
        self._triangles = triangles
        self._coordinate_system = coordinate_system

        self._transforms = []
        self._segmentations = ListOfNamed()

        self._normals = None
        self.normals = normals

        self._vertex_data = ListOfNamed()

    def __repr__(self) -> str:
        return 'Mesh: {} vertices, {} triangles'.format(self.nb_vertices,
                                                        self.nb_triangles)

    @property
    def coordinate_system(self) -> CoordinateSystem:
        """Returns the coordinate system of the mesh."""
        return self._coordinate_system

    @coordinate_system.setter
    def coordinate_system(self, coordinate_system):
        """Sets the coordinate system of the mesh"""

        if len(self._transforms) != 0:
            raise ValueError('The mesh contains transforms. The coordinate '
                             'system cannot be changed.')

        if not isinstance(coordinate_system, CoordinateSystem):
            raise TypeError('The coordinate system must be an instance of '
                            '{}, not {}.'
                            .format(CoordinateSystem, type(coordinate_system)))

        self._coordinate_system = coordinate_system

    @property
    def nb_triangles(self) -> int:
        """Returns the number of triangles of the mesh."""
        return len(self._triangles)

    @property
    def nb_vertices(self) -> int:
        """Returns the number of vertices of the mesh."""
        return len(self._vertices)

    @property
    def normals(self) -> Union[np.ndarray, None]:
        """Returns the normals of the mesh or None."""

        if self._normals is None:
            return None

        return self._normals.copy()

    @normals.setter
    def normals(self, normals: Sequence):
        """Sets the normals of the mesh."""

        if normals is None:
            self._normals = None
            return

        try:
            normals = np.array(normals, dtype=np.float64)
        except Exception:
            raise TypeError('\'normals\' must be convertible to a numpy '
                            'array of floats.')

        if normals.ndim != 2 or normals.shape[1] != 3:
            raise ValueError('\'normals\' must have a shape of (N, 3) not '
                             '{}.'.format(normals.shape))

        # The number of normals must match the number of vertices.
        if len(normals) != self.nb_vertices:
            raise ValueError('The number of normals must match the '
                             'number of vertices ({} != {}).'
                             .format(len(normals), self.nb_vertices))

        self._normals = normals

    @property
    def segmentations(self) -> ListOfNamed:
        """Returns the segmentations of the mesh."""
        return self._segmentations.copy()

    @property
    def transforms(self) -> List[AffineTransform]:
        """Returns the available transform to other coordinate systems."""
        return self._transforms.copy()

    @property
    def triangles(self) -> np.array:
        """Returns a copy of the mesh's triangles."""
        return self._triangles.copy()

    @property
    def vertex_data(self) -> ListOfNamed:
        """Returns the vertex data of the mesh"""
        return self._vertex_data.copy()

    @property
    def vertices(self) -> np.array:
        """Returns a copy of the mesh's vertices."""
        return self._vertices.copy()

    def add_segmentation(self, segmentation: 'Segmentation'):
        """Adds a segmentation to the mesh.

        Args:
            segmentation: The segmentation of the mesh to add.

        Raises:
            TypeError: If segmentation is not an instance of Segmentation.
            ValueError: If a segmentation with the same name already exists.

        """

        if not isinstance(segmentation, Segmentation):
            raise TypeError('\'segmentation\' must be an instance of '
                            'Segmentation, not {}.'
                            .format(type(segmentation)))

        if segmentation.name in [s.name for s in self._segmentations]:
            raise ValueError('A segmentation with the name {} already exists '
                             'in the mesh.'
                             .format(segmentation.name))

        self._segmentations.append(segmentation)

    def add_transform(self, transform: AffineTransform):
        """Adds a transform to a new coordinate system.

        Adds a new transform to the mesh that takes it from its current
        coordinate system to another.

        Args:
            transform: The affine transform from the current coordinate system
                to a new one.

        Raises:
            TypeError: If the transform is not an AffineTransform.
            ValueError: If a transform to the new coordinate system already
                exists or if the transform is to the current coordinate system.

        """

        if not isinstance(transform, AffineTransform):
            raise TypeError('\'transform\' must be an instance of '
                            'AffineTransform, not {}.'
                            .format(type(transform)))

        # Verify if the transform is to a new coordinate system.
        if transform.transform_coord_sys == self.coordinate_system:
            raise ValueError('A transform to the current coordinate system '
                             'cannot be added.')

        # Verify if there is already a transform to this coordiante system.
        targets = [t.transform_coord_sys for t in self._transforms]
        if transform.transform_coord_sys in targets:
            raise ValueError('A transform to {} already exists.'
                             .format(transform.transform_coord_sys))

        self._transforms.append(transform)

    def add_vertex_data(self, vertex_data: 'VertexData'):
        """Adds vertex data to the mesh

        Args:
            vertex_data: The vertex data to add. The name of this data must
                be distinct from the data already associated with the mesh. The
                length of the data must also match the number of vertices of
                the mesh.

        Raises:
            TypeError if vertex_data is not an instance of VertexData.
            ValueError if the mesh already has vertex data with the same name.
            ValueError if the length of the data does not match the number
                of vertices of the mesh.

        """

        if not isinstance(vertex_data, VertexData):
            raise TypeError('\'vertex_data\' must be an instance of '
                            'VertexData, not {}.'.format(type(vertex_data)))

        if vertex_data.name in [vd.name for vd in self._vertex_data]:
            raise ValueError('The mesh already contains vertex data with the '
                             'name {}.'.format(vertex_data.name))

        if vertex_data.nb_vertices != self.nb_vertices:
            raise ValueError('The number of vertices of the mesh does not '
                             'match the number of vertices of the data '
                             '({} != {}).'.format(vertex_data.nb_vertices,
                                                  self.nb_vertices))

        self._vertex_data.append(vertex_data)

    def transform_to(self, coordinate_system: CoordinateSystem):
        """Transforms the mesh to a another coordinate system.

        Transforms the mesh from its current coordinate system to another. A
        transform to the new coordinate system must have been added to the
        mesh with `add_transform`.

        Args:
             coordinate_system: The target coordinate system.

        Raises:
            ValueError: If no transform to the requested coordinate system
                exits.
        """

        # If we are already in this coordinate system, do nothing.
        if coordinate_system == self.coordinate_system:
            return

        # Get the requested transform.
        transform = next((t for t in self._transforms
                          if t.transform_coord_sys == coordinate_system), None)

        if transform is None:
            raise ValueError('No transform is available to transform to {}.'
                             .format(coordinate_system))

        # Apply the transform to the vertices.
        homogeneous_vertices = np.hstack((self._vertices,
                                          np.ones((self.nb_vertices, 1))))
        new_vertices = np.dot(transform.affine, homogeneous_vertices.T)
        self._vertices = new_vertices[:3, :].T

        # Remove the transform form the list because it is no longer valid.
        self._transforms.remove(transform)

        # Apply inverse of the transform to the other transform so they still
        # project to the correct coordinate system.
        inv = np.linalg.inv(transform.affine)

        def project_transform(t):
            return AffineTransform(t.transform_coord_sys,
                                   np.dot(inv, t.affine))
        self._transforms = [project_transform(t) for t in self._transforms]

        # Add the inverse of the transform to go back to the original
        # coordinate system and the change the current coordinate system.
        new_transform = AffineTransform(self.coordinate_system, inv)
        self._coordinate_system = coordinate_system
        self.add_transform(new_transform)


class Segmentation(Named):

    def __init__(self, name: str, keys: Sequence[int]):
        """Segmentation of a mesh using labels.

        A segmentation of a mesh represented using a list of integers,
        one for each vertex.

        Args:
            name: The name of the segmentation.
            keys: The array of labels of the segmentation, one for each
                voxel. Must be a sequence convertible to a numpy array of
                integers with a shape of (N,).

        Raises:
            TypeError: If the labels cannot be converted to an array of
                integers.
            ValueError: If the labels cannot be converted to an array with a
                shape of (N,).
        """

        super().__init__(name)

        # The keys must be convertible to a numpy array of integers.
        try:
            keys = np.array(keys, np.int32)
        except Exception:
            raise TypeError('\'keys\' must be convertible to a numpy array '
                            'of integers.')

        # The shape of the keys must be (N,).
        if keys.ndim != 1:
            raise ValueError('\'keys\' must have a shape of (N,), not {}.'
                             .format(keys.shape))

        self._keys = keys
        self._labels = {}

    @property
    def keys(self) -> np.array:
        """Returns the key of each vertex of the segmentation."""
        return self._keys.copy()

    @property
    def labels(self) -> dict:
        """Returns a dict of labels."""
        return self._labels.copy()

    def add_label(self, key: int, label: Label):
        """Adds a label to the segmentation.

        Adds a label to the segmentation by associating it with a key.
        Because each vertex is also assigned a key, this creates a mapping
        between vertices and labels.

        Args:
            key: The key of the label.
            label: The label data of the label.

        """

        self._labels[key] = label


class VertexData(Named):
    def __init__(self, name: str, data: Sequence):
        """Vertex data of a mesh

        The vertex data class represents arbitrary data about the vertices
        of a mesh. The data must be convertible to an array of float with a
        shape of (N, M) where N is the number of vertices of the mesh.

        Args:
            name: The name of the vertex data.
            data: The data of each vertex.

        Raises:
            TypeError if the data cannot be converted to a numpy array of
                floats.
            ValueError if the data array does not have a shape of (N, M).

        """

        super().__init__(name)

        self._data = None
        self.data = data

    @property
    def data(self) -> np.ndarray:
        """Returns the data for each vertex"""
        return self._data

    @data.setter
    def data(self, data: Sequence):
        """Sets the data for each vertex

        Modifies the data associated with each vertex. The length of the
        data must match the length of the previous data to prevent changing
        the number of vertices.

        Args:
            data: The data of each vertex. Must be a Sequence convertible to a
                numpy array of floats.

        Raises:
            TypeError if the data cannot be converted to a numpy array of
                floats.
            ValueError if the data array does not have a shape of (N, M).
            ValueError if the length of the data does not match the length
                of the previous data.

        """

        try:
            data = np.array(data, dtype=np.float64)
        except Exception:
            raise TypeError('\'data\' must be convertible to a numpy '
                            'array of floats.')

        if data.ndim == 1:
            data = data[:, None]

        elif data.ndim != 2:
            raise ValueError('\'data\' must have two dimensions, not {}.'
                             .format(data.ndim))

        if self.data is not None and len(data) != self.nb_vertices:
            raise ValueError('The length of \'data\' does not match the '
                             'number of vertices of the previous data '
                             '({} != {}).'.format(len(data), self.nb_vertices))

        self._data = data

    @property
    def nb_vertices(self):
        """Returns the number of vertices associated with the data"""
        return len(self.data)



import warnings
from typing import Union

import nibabel as nib

from nimesh import AffineTransform, CoordinateSystem, Mesh, Segmentation
from nimesh.core import VertexData


def load(filename: str) -> Mesh:
    """Loads a mesh from a GifTI file

    Loads the vertices and triangles of a mesh in the GifTI file format. If
    a transform or segmentation is contained in the file, they will also be
    loaded.

    Args:
        filename: The name of the GifTI file to load.

    Returns:
        mesh: The mesh loaded from the supplied file.

    Raises:
        ValueError: If the file does not contains a vertex and triangle array.

    Warnings:
        RuntimeWarning: If the file contains more than one vertex or
            triangle array.

    """

    gii = nib.load(filename)

    # Get the vertices array. If there is more than one, warn the user and
    # use the first one.
    vertex_arrays = gii.get_arrays_from_intent('NIFTI_INTENT_POINTSET')

    if len(vertex_arrays) == 0:
        raise ValueError('The file {} does not contain any vertex data.'
                         .format(filename))

    elif len(vertex_arrays) > 1:
        warnings.warn('The file {} contains more than one vertex array. The '
                      'first one was used. Proceed with caution.'
                      .format(filename),
                      RuntimeWarning)
    vertex_array = vertex_arrays[0]
    vertices = vertex_array.data

    # Get the coordinate system from the metadata.
    if 'cs' in vertex_array.metadata:
        coordinate_system = CoordinateSystem(int(vertex_array.metadata['cs']))
    else:
        coordinate_system = _convert_nifti_code_to_coord_sys(
            vertex_array.coordsys.dataspace)

    # Get the triangles array. If there is more than one, warn the user and
    # use the first one.
    triangle_arrays = gii.get_arrays_from_intent('NIFTI_INTENT_TRIANGLE')

    if len(triangle_arrays) == 0:
        raise ValueError('The file {} does not contain any triangles.'
                         .format(filename))

    elif len(triangle_arrays) > 1:
        warnings.warn('The file {} contains more than one triangle array. '
                      'The first one was used. Proceed with caution.'
                      .format(filename),
                      RuntimeWarning)

    triangles = triangle_arrays[0].data

    # Get the normals array if it exists.
    normals_arrays = gii.get_arrays_from_intent('NIFTI_INTENT_VECTOR')
    if len(normals_arrays) > 0:
        if len(normals_arrays) != 1:
            warnings.warn('The file {} contains more than one vector array. '
                          'The first one was interpreted as mesh normals. '
                          'Proceed with caution.'.format(filename))

        normals = normals_arrays[0].data

    else:
        normals = None

    # Create the mesh.
    mesh = Mesh(vertices, triangles, coordinate_system, normals)

    # Add the transform if one was saved.
    transform = _get_transform_from_vertex_array(vertex_array)
    if transform is not None:
        mesh.add_transform(transform)

    # Add the segmentation if one was saved.
    segmentation = _create_segmentation_from_gii(gii)
    if segmentation is not None:
        mesh.add_segmentation(segmentation)

    # Add the vertex data if any was saved.
    vertex_data_arrays = gii.get_arrays_from_intent('NIFTI_INTENT_ESTIMATE')
    for vertex_data_array in vertex_data_arrays:

        if 'name' not in vertex_data_array.meta.metadata:
            warnings.warn('The file {} contains a vertex data array without '
                          'a name. It was ignored.'.format(filename))
            continue

        vertex_data = VertexData(vertex_data_array.meta.metadata['name'],
                                 vertex_data_array.data)
        mesh.add_vertex_data(vertex_data)

    return mesh


def load_segmentation(filename: str) -> Segmentation:
    """Loads a segmentation from a GifTI file.

    Loads the segmentation of a mesh from a GifTI file without loading the
    mesh data.

    Args:
        filename: The name of the file from which to load the segmentation.

    Returns:
        segmentation: The segmentation loaded from the GifTI file.

    Raises:
        ValueError: If the file does not contain a segmentation.

    """

    gii = nib.load(filename)

    segmentation = _create_segmentation_from_gii(gii)

    if segmentation is None:
        raise ValueError('The file {} does not contain any segmentation data.'
                         .format(filename))

    return segmentation


def save(filename: str, mesh: Mesh,
         anatomical_structure_primary: str = 'Cortex',
         anatomical_structure_secondary: str = 'GrayWhite',
         geometric_type: str = 'Anatomical') -> None:
    """Save a mesh to a GifTI file.

    Saves the mesh and its metadata to a GifTI file.

    Args:
        filename: The name of the file were the mesh will be saved. If it
            exists, it will be overwritten.
        mesh: The mesh to save.
        anatomical_structure_primary (optional): The tag of the primary
            anatomical structure for workbench.
        anatomical_structure_secondary (optional): The tag of the secondary
            anatomical structure for workbench.
        geometric_type (optional): The geometric type for workbench.

    """

    gii = nib.gifti.GiftiImage()

    # The GifTI file format only allows saving one segmentation. If the mesh
    # contains more that one, warn the user and save only the first one.
    segmentations = mesh.segmentations
    if len(segmentations) > 1:
        warnings.warn('The mesh has more that one segmentation, but GifTI '
                      'only supports one segmentation per file. Only the '
                      'first segmentation will be saved.')

    if len(segmentations) != 0:
        add_segmentation_to_gii(segmentations[0], gii)

    # The coordinate system is saved in the metadata of the vertex array.
    # Because it is possible to save the coordinate system of the points
    # without having a transform, we add it to the metadata.
    meta = {
        'cs': str(mesh.coordinate_system.value)
    }

    # For now, the GifTI implementation of nibabel seems to support only a
    # single transform. If the mesh has more than one, warn the user and
    # save the first one.
    transforms = mesh.transforms
    if len(transforms) > 1:
        warnings.warn('The mesh has more than one transform but GifTI only '
                      'supports a single transform. Only the first one will '
                      'be saved.')

    if len(transforms) != 0:
        transform = transforms[0]
        coordinate_system = nib.gifti.GiftiCoordSystem(
            _convert_coord_sys_to_nifti_code(mesh.coordinate_system),
            _convert_coord_sys_to_nifti_code(transform.transform_coord_sys),
            transform.affine)
        meta['tcs'] = str(transform.transform_coord_sys.value)
    else:
        coordinate_system = None

    vertices_array = nib.gifti.GiftiDataArray(
        mesh.vertices.astype('f4'),
        intent='NIFTI_INTENT_POINTSET',
        datatype='NIFTI_TYPE_FLOAT32',
        meta=meta,
        coordsys=coordinate_system
    )

    # Add metadata for Workbench. In a surface file, the meta data of the
    # structure is in the meta of the point set array.
    vertices_array.meta.data.append(
        nib.gifti.GiftiNVPairs('AnatomicalStructurePrimary',
                               anatomical_structure_primary))
    vertices_array.meta.data.append(
        nib.gifti.GiftiNVPairs('AnatomicalStructureSecondary',
                               anatomical_structure_secondary))
    vertices_array.meta.data.append(
        nib.gifti.GiftiNVPairs('GeometricType', geometric_type))

    gii.add_gifti_data_array(vertices_array)

    triangles_array = nib.gifti.GiftiDataArray(
        mesh.triangles.astype('i4'),
        intent='NIFTI_INTENT_TRIANGLE',
        datatype='NIFTI_TYPE_INT32')
    gii.add_gifti_data_array(triangles_array)

    # Save the normals if they exist.
    if mesh.normals is not None:
        normals_array = nib.gifti.GiftiDataArray(
            mesh.normals.astype('f4'),
            intent='NIFTI_INTENT_VECTOR',
            datatype='NIFTI_TYPE_FLOAT32')
        gii.add_gifti_data_array(normals_array)

    # Save the vertex data if there is any.
    for vertex_data in mesh.vertex_data:

        vertex_data_array = nib.gifti.GiftiDataArray(
            vertex_data.data.astype('f4'),
            intent='NIFTI_INTENT_ESTIMATE',
            datatype='NIFTI_TYPE_FLOAT32',
            meta={'name': vertex_data.name})
        gii.add_gifti_data_array(vertex_data_array)

    nib.save(gii, filename)


def save_segmentation(filename: str,
                      segmentation: Segmentation,
                      anatomical_structure_primary: str = 'Cortex',
                      anatomical_structure_secondary: str = 'GrayWhite',
                      geometric_type: str = 'Anatomical') -> None:
    """Save a segmentation to a GifTI image.

    Save a the segmentation of a mesh into a GifTI file without saving the
    mesh itself.

    Args:
        filename: The name of the file where the segmentation is saved. If
            it exists, it will be overwritten.
        segmentation: The segmentation to save.
        anatomical_structure_primary: The name of the primary anatomical
            structure the segmentation refers to. For workbench.
        anatomical_structure_secondary: The name of the secondary anatomical
            structure the segmentation refers to. For workbench.
        geometric_type: The geometric type of the structure. For workbench.

    """

    gii = nib.gifti.GiftiImage()

    # In a label file, the meta data of the structure is in the
    # meta of the file.
    gii.meta.data.append(
        nib.gifti.GiftiNVPairs('AnatomicalStructurePrimary',
                               anatomical_structure_primary))
    gii.meta.data.append(
        nib.gifti.GiftiNVPairs('AnatomicalStructureSecondary',
                               anatomical_structure_secondary))
    gii.meta.data.append(
        nib.gifti.GiftiNVPairs('GeometricType',
                               geometric_type))

    add_segmentation_to_gii(segmentation, gii)

    nib.save(gii, filename)


def add_segmentation_to_gii(segmentation: Segmentation,
                            gii: nib.gifti.GiftiImage):
    """Adds a segmentation to a nibabel GifTI object.

    Add the segmentation data to a nibabel GifTI object by adding a new data
    array.

    Args:
        segmentation: The segmentation to add to the GifTI object.
        gii: The GifTI object where the segmentation is added.

    """

    # Save the name of the segmentation in the metadata of the array.
    meta = nib.gifti.GiftiMetaData.from_dict({'name': segmentation.name})

    label_array = nib.gifti.GiftiDataArray(
        segmentation.keys,
        intent='NIFTI_INTENT_LABEL',
        datatype='NIFTI_TYPE_INT32',
        meta=meta)

    gii.add_gifti_data_array(label_array)


def _convert_coord_sys_to_nifti_code(coord_sys: CoordinateSystem) -> int:
    """Converts coordinate systems to NIfTI xform codes"""

    convert = {
        CoordinateSystem.UNKNOWN: 0,
        CoordinateSystem.SCANNER: 1,
        CoordinateSystem.RAS: 2,
        CoordinateSystem.LPS: 2,
        CoordinateSystem.VOXEL: 0,
        CoordinateSystem.MNI: 4,
        CoordinateSystem.TALAIRACH: 3,
    }

    return convert[coord_sys]


def _convert_nifti_code_to_coord_sys(code: int) -> CoordinateSystem:
    """Converts a NIfTI xform code to a coordinate system"""

    convert = {
        0: CoordinateSystem.UNKNOWN,
        1: CoordinateSystem.SCANNER,
        2: CoordinateSystem.UNKNOWN,
        3: CoordinateSystem.TALAIRACH,
        4: CoordinateSystem.MNI,
    }

    return convert[code]


def _create_segmentation_from_gii(gii) -> Segmentation:
    """Creates a segmentation from a nibabel GifTI object.

    Creates a new segmentation from the data array of a nibabel GifTI
    object. If none of the data arrays have an intent of NIFTI_INTENT_LABEL,
    None is returned.

    Args:
        gii: The nibabel GifTI object from which to create the segmentation.

    Returns:
        segmentation: The loaded segmentation or None.

    """

    # Get the labels array. If there is more than one, warn the user and
    # use the first one.
    label_arrays = gii.get_arrays_from_intent('NIFTI_INTENT_LABEL')

    if len(label_arrays) == 0:
        return None

    elif len(label_arrays) > 1:
        warnings.warn('The file contains more than one label array. The '
                      'first one was used. Proceed with caution.',
                      RuntimeWarning)
    label_array = label_arrays[0]
    labels = label_array.data
    name = label_array.metadata['name']

    return Segmentation(name, labels)


def _get_transform_from_vertex_array(vertex_array)\
        -> Union[AffineTransform, None]:
    """Gets the affine transform from a GIfTI vertex array"""

    # GIfTI always contain a transform.
    gifti_transform = vertex_array.coordsys

    # If the metadata of the vertex array contains the key 'tcs',
    # use it to create the transform. If is doesn't, only add the
    # transform if the dataspace and xformspace are not identical.
    if 'tcs' in vertex_array.metadata:
        transform = AffineTransform(
            CoordinateSystem(int(vertex_array.metadata['tcs'])),
            gifti_transform.xform)

    elif gifti_transform.dataspace != gifti_transform.xformspace:
        transform = AffineTransform(
            _convert_nifti_code_to_coord_sys(gifti_transform.xformspace),
            gifti_transform.xform)

    else:
        transform = None

    return transform



import os
import warnings

import nibabel as nib
import numpy as np
from nimesh.core import AffineTransform, CoordinateSystem, Mesh
from nimesh.core import Label, Segmentation
from nimesh import io


def from_freesurfer(subject_directory: str,
                    hemisphere: str,
                    surface: str):
    """Loads a mesh from a FreeSurfer subject directory

    Args:
        subject_directory: The path to the subject's directory.
        hemisphere: The hemisphere to load. Must be either 'lh' or 'rh'.
        surface: The surface to load. Must be 'pial', 'white', or 'inflated'.

    Raises:
        ValueError: If `hemisphere` or `surface` are not valid strings.

    """

    # Validate the hemisphere and the surface names.
    if hemisphere not in ('rh', 'lh'):
        raise ValueError('\'hemisphere\' must be \'lh\' of \'rh\', not \'{}\'.'
                         .format(hemisphere))

    if surface not in ('pial', 'white', 'inflated', 'smoothwm'):
        raise ValueError('\'surface\' must be \'pial\', \'white\', '
                         'or \'inflated\', not \'{}\'.'
                         .format(surface))

    # Load the requested surface.
    surface_file = os.path.join(subject_directory, 'surf',
                                hemisphere + '.' + surface)
    mesh = io.freesurfer.load(surface_file)

    # Get the coordinate system information from the MRI image.
    nii = nib.load(os.path.join(subject_directory, 'mri', 'rawavg.mgz'))
    codes = nib.aff2axcodes(nii.affine)

    if codes == ('L', 'P', 'S'):
        coordinate_system = CoordinateSystem.LPS
    elif codes == ('R', 'A', 'S'):
        coordinate_system = CoordinateSystem.RAS
    else:
        coordinate_system = CoordinateSystem.UNKNOWN
        warnings.warn('The coordinate system could not be identified from '
                      'the MRI (rawavg.mgz). The transform will not be set.')

    mesh.coordinate_system = coordinate_system

    # Add the transform to voxel space if the coordinate system is known.
    if coordinate_system != CoordinateSystem.UNKNOWN:
        transform = AffineTransform(CoordinateSystem.VOXEL,
                                    np.linalg.inv(nii.affine))
        mesh.add_transform(transform)

    # Load the annotations.
    desikan_file = os.path.join(subject_directory, 'label',
                                hemisphere + '.aparc.annot')
    mesh.add_segmentation(io.freesurfer.load_segmentation(desikan_file))
    dkt_file = os.path.join(subject_directory, 'label',
                            hemisphere + '.aparc.DKTatlas.annot')
    mesh.add_segmentation(io.freesurfer.load_segmentation(dkt_file))
    destrieux_file = os.path.join(subject_directory, 'label',
                                  hemisphere + '.aparc.a2009s.annot')
    mesh.add_segmentation(io.freesurfer.load_segmentation(destrieux_file))

    return mesh



from typing import Union


class Named(object):

    def __init__(self, name):
        """Mixin for objects with a name."""
        self._name = name

    @property
    def name(self):
        """Returns the name of the object."""
        return self._name


class ListOfNamed(list):
    """Represents a list of items that have names."""

    def __getitem__(self, item: Union[str, int, slice]) -> Union[Named, None]:
        """Allows getting items by their name"""

        if isinstance(item, str):
            return next((o for o in self if o.name == item), None)
        else:
            return super().__getitem__(item)

    def copy(self) -> 'ListOfNamed':
        """Copies a list of named object."""
        return ListOfNamed(self)



import os
import warnings

import nibabel.freesurfer as nibfs
import numpy as np

from nimesh import CoordinateSystem, Label, Mesh, Segmentation


def load(surface: str, annotation: str = None) -> Mesh:
    """Loads a FreeSurfer surface mesh.

    Loads a mesh from a FreeSurfer geometry file, optionally loading
    annotations.

    Args:
        surface: The FreeSurfer geometry file, e.g. lh.pial or rh.white.
        annotation (optional): The annotation file to load, for example
            lh.aparc.DKTatlas.annot or lh.aparc.annot.

    Returns:
        mesh: The loaded mesh.

    """

    # Load the vertices and remove the CRAS translation if it exists.
    vertices, triangles, meta, info = nibfs.read_geometry(surface, True, True)
    if 'cras' in meta:
        vertices += meta['cras']

    mesh = Mesh(vertices, triangles)

    # Load the annotations, if requested.
    if annotation is not None:
        segmentation = load_segmentation(annotation)
        mesh.add_segmentation(segmentation)

    return mesh


def load_segmentation(filename: str) -> Segmentation:
    """Loads a segmentation from a FreeSurfer annot file.

    Loads the a segmentation from a FreeSurfer annotation file. Because annot
    files do not have a segmentation name, the file name is used instead.

    Args:
        filename: The name of the annot file that is loaded.

    Returns:
        segmentation: A segmentation whose name is `filename`.

    """

    keys, ctab, names = nibfs.read_annot(filename)
    segmentation = Segmentation(os.path.basename(filename), keys)

    def rgbt2rgba(color):
        return color[0], color[1], color[2], 255

    # Add the label names.
    for i, (color, name) in enumerate(zip(ctab, names)):
        segmentation.add_label(i, Label(name.decode(), rgbt2rgba(color[:4])))

    return segmentation


def save(filename: str, mesh: Mesh):
    """Saves the vertices and triangles of a mesh to a FreeSurfer file."""

    # Convert the mesh to RAS.
    if mesh.coordinate_system != CoordinateSystem.RAS:
        available_transforms = [t.transform_coord_sys for t in mesh.transforms]
        if CoordinateSystem.RAS not in available_transforms:
            raise ValueError('To save a mesh in the FreeSurfer format, '
                             'it must be in the RAS coordinate system or '
                             'have a transform to RAS.')

        warnings.warn('The mesh is not in RAS, transforming it to RAS.')
        mesh.transform_to(CoordinateSystem.RAS)

    # Generate fake metatdata for the geometry file.
    meta = {
        'head': np.array([2, 0, 20], dtype=np.int32),
        'valid': '1 # volume info valid',
        'filename': filename,
        'volume': np.array([256, 256, 256]),
        'voxelsize': np.array([1, 1, 1]),
        'xras': [1, 0, 0],
        'yras': [0, 1, 0],
        'zras': [0, 0, 1],
        'cras': [0, 0, 0],
    }

    nibfs.write_geometry(filename, mesh.vertices, mesh.triangles, True, meta)


def save_segmentation(filename: str, segmentation: Segmentation):
    """Saves a segmentation to a FreeSurfer annot file.

    Saves a segmentation to a FreeSurfer annotation file. To be saved
    correctly, each label needs to have a distinct color. The current
    implementation relies on nibabel and does not save/load alpha values
    correctly.

    Args:
        filename: The name of the annotation file where the segmentation will
            be saved.
        segmentation: The segmentation to save.

    Raises:
        ValueError if the segmentation has labels with duplicate colors.
            This is a limitation imposed by the FreeSurfer annotation format.

    """

    # nibabel does not appear to save the alpha value and always returns
    # 255. Warn the user that alpha values will be lost if they are not all
    # 255.
    alphas = {label.color[-1] for label in segmentation.labels.values()}
    if np.any(alphas != 255):
        warnings.warn('Alpha values for labels currently cannot be saved in '
                      'the FreeSurfer annotation file format. They will be '
                      'lost.',
                      RuntimeWarning)

    # To save a segmentation to a FreeSurfer annotation file, all labels must
    # have a distinct color.
    colors = {label.color for label in segmentation.labels.values()}
    if len(colors) != len(segmentation.labels):
        raise ValueError('To save a segmentation to a FreeSurfer annotation '
                         'file, all labels must have a distinct color.')

    def color2annot(color):
        return color[0] + \
               color[1] * 256 + \
               color[2] * 256 ** 2 + \
               (255 - color[3]) * 256 ** 3

    def rgba2rgbt(color):
        """Change alpha to transparency in color"""
        return color[0], color[1], color[2], 255 - color[3]

    keys = segmentation.keys
    ctab = np.array([(*rgba2rgbt(label.color), color2annot(label.color))
                     for key, label in segmentation.labels.items()])
    names = [label.name for label in segmentation.labels.values()]
    nibfs.write_annot(filename, keys, ctab, names)



from setuptools import setup

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='nimesh',
    version='0.0.0',
    packages=['nimesh', 'nimesh.io'],
    url='https://github.com/sdeslauriers/nimesh',
    license='MIT',
    author='Samuel Deslauriers-Gauthier',
    author_email='sam.deslauriers@gmail.com',
    description='A Python package to manipulate neuroimaging triangular '
                'meshes.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['numpy', 'nibabel'],
    extras_requires={
        'visualization': ['vtk'],
    },
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ),
)



import numpy as np


def compute_normals(vertices: np.ndarray, faces: np.ndarray) -> np.ndarray:
    """ Computes the vertex normals of a mesh

    The normal of a vertex is the average normal of all faces that include
    the vertex.

    Args:
        vertices: A 2D numpy array with a shape of (N, 3)
            where N is the number of vertices.
        faces: A 2D numpy array with a shape of (M, 3)
            where M is the number of faces.

    Returns:
        normals: A 2D numpy array with a shape of (N, 3) that
            contains the normal for each vertex.

    Raises:
        ValueError if a vertex has a normal with a norm of 0.

    """

    normals = []
    for i in range(len(vertices)):

        # Get all faces which involve the vertex.
        contains_vertex = np.any(faces == i, axis=1)
        vertex_faces = faces[contains_vertex, :]

        normal = np.zeros((3,))
        for face in vertex_faces:
            normal += np.cross(vertices[face[1]] - vertices[face[0]],
                               vertices[face[2]] - vertices[face[0]])

        # Normalize the normal.
        norm = np.linalg.norm(normal)
        if norm == 0:
            raise ValueError('A vertex has a 0 norm normal')

        normal /= norm
        normals.append(normal)

    return np.array(normals)



import os.path

from nimesh import Mesh
from nimesh.io import gifti
from nimesh.io import freesurfer


def load(filename: str) -> Mesh:
    """Loads a mesh from a file.

    The file format is detected from the filename extension. For now, the only
    supported file format is GifTI (.gii).

    Args:
        filename: The name of the file to load. Must have a .gii extension.

    Returns:
        mesh: The mesh contained in the supplied file.

    Raises:
        ValueError: If the file extension is not supported.

    """

    extension = os.path.splitext(filename)[1]

    if extension == '.gii':
        mesh = gifti.load(filename)

    else:
        raise ValueError('Unknown file format {}.'.format(extension))

    return mesh


def save(filename: str, mesh: Mesh) -> None:
    """ Saves a mesh to a file

    The output format depends on the extension of the supplied filename. For
    now, the only supported file format is GifTI (.gii).

    Args:
        filename: The name of the file were the mesh will be saved. If it
            exists, it will be overwritten.
        mesh: The mesh to save.

    Raises:
        ValueError: If the file extension is not supported.

    """

    extension = os.path.splitext(filename)[1]

    if extension == '.gii':
        gifti.save(filename, mesh)

    else:
        raise ValueError('Unknown file format {}.'.format(extension))
